unit uPlugin;

interface
uses Windows, DVSTUtils, uEditor, DAEffect, DAEffectX,
     DAudioEffect, DAudioEffectX, WaveIOX;

const
 kID = 'TBWT';
 kChannelID = 'TBWT';
 kEffectName = 'WavTester';
 kProduct = 'WavTester';
 kVendor = 'Tobybear';
 kIsSynth = true;
 kNumInputs = 2;
 kNumOutputs = 2;
 kCanMono = true;
 kCanReplacing = true;

 kpar01 = 0; // volume
 kpar02 = 1; // panning
 kpar03 = 2; // start playing
 kpar04 = 3; // loop start point
 kpar05 = 4; // inerpolation
 kpar06 = 5; // loop end point
 kpar07 = 6; // pitch
 kpar08 = 7; // stop playing

 kwavload = 111;

 kNumParams = 8;
 kNumPrograms = 1;

 type APluginEditor = class;
 APluginProgram = class
 private
  fpar01, fpar02, fpar03, fpar04,
  fpar05, fpar06, fpar07, fpar08:single;
  name: array[0..50] of char;
 public
  constructor Create;
 end;

 APlugin = class(AudioEffectX)
 private
  host1, host2:string;
  nvol: single;
  pbuf: pointer;
  pf: psingle;
  cnt2, size, sr, ch: longint;
  cnt: double;
  fpar01, fpar02, fpar03, fpar04,
  fpar05, fpar06, fpar07, fpar08: single;
  done: boolean;
 public
  pmode: longint;
  programs: array[0 .. kNumPrograms - 1] of APluginProgram;
  vu: single;
  function GetPluginEditor: APluginEditor;
  constructor CreateAPlugin(audioMaster: TAudioMasterCallbackFunc); virtual;
  destructor Destroy; override;
  procedure DoProcess(var i1, i2: single);
  procedure process(inputs, outputs: PPSingle; sampleframes: Longint); override;
  procedure processReplacing(inputs, outputs: PPSingle; sampleframes: Longint); override;
  procedure setProgram(aProgram: Longint); override;
  function canDo(text: pchar): longint; override;
  procedure setProgramName(name: PChar); override;
  procedure getProgramName(name: PChar); override;
  procedure setParameter(index: Longint; value: Single); override;
  function getParameter(index: Longint): Single; override;
  function getVu: Single; override;
  function processEvents(ev: PVSTEvents): longint; override;
  procedure resume; override;
  procedure suspend; override;
  procedure LoadFile(fn: string);
  procedure getParameterLabel(index: longint; text: pchar); override;
  procedure getParameterDisplay(index: longint; text: pchar); override;
  procedure getParameterName(index: longint; text: pchar); override;
  function getOutputProperties(index: longint; properties: PVstPinProperties): boolean; override;
  function getProgramNameIndexed(category, index: longint; text: pchar): boolean; override;
  function getEffectName(name: pchar): boolean; override;
  function getVendorString(text: pchar): boolean; override;
  function getProductString(text: pchar): boolean; override;
  function getVendorVersion: longint; override;
  property PluginEditor: APluginEditor read GetPluginEditor;
 end;

 APluginEditor = class(AEffEditor)
 private
  r: ERect;
  useCount: Longint;
  Editor: TPluginEditorWindow;
  systemWindow: HWnd;
  function GetPlugin: APlugin;
 public
  constructor Create(effect: AudioEffect); override;
  destructor Destroy; override;
  function getRect(var rect: PERect): Longint; override;
  function open(ptr: Pointer): Longint; override;
  procedure close; override;
  procedure idle; override;
  procedure update; override;
  property Plugin: APlugin read GetPlugin;
 end;

implementation
uses graphics, SysUtils, dialogs, mmsystem, classes;

function APlugin.canDo(text: pchar): longint;
begin
 Result := -1;
 if StrComp(text, 'receiveVstEvents') = 0 then Result := 1;
 if StrComp(text, 'receiveVstMidiEvent') = 0 then Result := 1;
end;

constructor APluginProgram.Create;
begin
 inherited Create;
 // Here the parameters for a program are inititialized
 fpar01 := 1;
 fpar02 := 0.5;
 fpar03 := 0;
 fpar04 := 0;
 fpar05 := 1;
 fpar06 := 1;
 fpar07 := 0.5;
 fpar08 := 0;
 StrCopy(name, 'Init'); // set program name
end;

constructor APlugin.CreateAPlugin(audioMaster: TAudioMasterCallbackFunc);
var i: integer;
    t: array[0..50] of char;
begin
 inherited Create(audioMaster, knumprograms, kNumParams);
 pbuf := nil;
 pf := nil;
 pmode := 0;
 cnt := 0;
 cnt2 := 0;

 // create the programs
 for i := kNumPrograms - 1 downto 0 do programs[i] := APluginProgram.Create;

 editor := APluginEditor.Create(Self);
 randomize; // important if you use the random() function!
 suspend;

 hasVu(false);
 setNumInputs(KNumInputs);
 setNumOutputs(KNumOutputs);
 canMono(KCanMono);
 canProcessReplacing(KCanReplacing);
 isSynth(KIsSynth);
 setUniqueID(FourCharToLong(kID[1], kID[2], kID[3], kID[4]));

 // initial program names
 StrCopy(programs[0].name, 'Init Preset');
 curprogram := 0;
 setProgram(0);
 gethostvendorstring(t); host1 := strpas(t);
 gethostproductstring(t); host2 := strpas(t);
end;

destructor APlugin.Destroy;
var i:integer;
begin
 inherited Destroy;
 if assigned(pbuf) then freemem(pbuf);
 for i := 0 to kNumPrograms - 1 do
 begin
  programs[i].Free;
  programs[i] := nil;
 end;
end;

procedure APlugin.setProgram(aProgram: Longint);
begin
 if (aprogram < 0) or (aprogram > knumprograms - 1) then exit;
 curProgram := aProgram;
 SetParameter(kpar01, programs[curprogram].fpar01);
 SetParameter(kpar02, programs[curprogram].fpar02);
 SetParameter(kpar03, programs[curprogram].fpar03);
 SetParameter(kpar04, programs[curprogram].fpar04);
 SetParameter(kpar05, programs[curprogram].fpar05);
 SetParameter(kpar06, programs[curprogram].fpar06);
 SetParameter(kpar07, programs[curprogram].fpar07);
 SetParameter(kpar08, programs[curprogram].fpar08);
 if done then exit;
 if assigned(editor) then editor.update;
end;

procedure APlugin.setProgramName(name: PChar);
begin
 StrCopy(programs[curProgram].name, name);
 if done then exit;
 if assigned(editor) then editor.update;
end;

procedure APlugin.getProgramName(name: PChar);
begin
 StrCopy(name, programs[curProgram].name);
end;

procedure APlugin.suspend;
begin
end;

function APlugin.getVu: Single;
var cvu: Single;
begin
 cvu := vu;
 vu := 0;
 Result := cvu;
end;

procedure APlugin.LoadFile(fn: string);
var old: integer;
begin
 old := pmode;
 pmode := 0;
 if assigned(pbuf) then
 begin
  freemem(pbuf);
  pbuf := nil;
 end;
 filemode := 0;

 // load file
 pbuf := LoadWAVFile(fn, sr, ch, size);
 pf := pbuf;

 // reset all counters:
 cnt := 0;
 cnt2 := 0;
 pmode := old;
end;

procedure APlugin.setParameter(index: Longint; value: Single);
var cx, c, stp: longint;
    ps: psingle;
begin
 if (value > 1) then value := 1 else if (value < 0) then value := 0;

 case index of
  kwavload:
  begin
   if assigned(editor) then
   if assigned(plugineditor) then
   if assigned(plugineditor.editor) then
   begin
    loadfile(plugineditor.Editor.OpenDialogSource.filename);
    stp := size div 600; c := 0; cx := 0;
    plugineditor.editor.wdisplay.Canvas.brush.color := clblack;
    plugineditor.editor.wdisplay.Canvas.FillRect(rect(0, 0, 600, 135));
    plugineditor.editor.wdisplay.Canvas.pen.color := clred;
    plugineditor.editor.wdisplay.Canvas.moveto(0, 65);
    repeat
     ps := psingle(longint(pbuf) + sizeof(single) * c);
     plugineditor.editor.wdisplay.Canvas.lineto(cx,65 - round(ps^ * 60));
     c := c + stp;
     cx := cx + 1;
    until cx >= 599;
   end;
  end;
  kpar01: begin fpar01 := value; programs[curprogram].fpar01 := value end;
  kpar02: begin fpar02 := value; programs[curprogram].fpar02 := value end;
  kpar03: begin
   fpar03 := value; programs[curprogram].fpar03 := value;
   if value > 0.9 then pmode := 1;
  end;
  kpar04: begin fpar04 := value; programs[curprogram].fpar04 := value end;
  kpar05: begin fpar05 := value; programs[curprogram].fpar05 := value end;
  kpar06: begin fpar06 := value; programs[curprogram].fpar06 := value end;
  kpar07: begin fpar07 := value;programs[curprogram].fpar07 := value end;
  kpar08: begin fpar08 := value;programs[curprogram].fpar08 := value;
   if value > 0.9 then pmode := 0;
  end;
 end;
 if assigned(editor) then editor.update;
end;

function APlugin.getParameter(index: Longint): Single;
var j: single;
begin
 case index of
  kpar01: j := fpar01;
  kpar02: j := fpar02;
  kpar03: j := fpar03;
  kpar04: j := fpar04;
  kpar05: j := fpar05;
  kpar06: j := fpar06;
  kpar07: j := fpar07;
  kpar08: j := fpar08;
 else
  j := 0;
 end;
 Result := j;
end;

procedure APlugin.DoProcess(var i1, i2: single);
var o1, o2: single;
    ls, le: longint;
    next, next2, pp: psingle;
begin
 if (not assigned(pf)) // if buffer is empty (no file loaded)
 or (pmod e= 0)          // or "play" not activated
 then begin            // then output silence
  o1 := 0;
  o2 := 0;
 end else
 begin
  o1 := pf^;
  if ch = 2 then // stereo?
  begin
   pp := psingle(longint(pf) + 4);
   o2 := pp^;
   next := psingle(longint(pf) + 8);
   next2 := psingle(longint(pf) + 12);
   o2 := o2 * (1 - cnt) + cnt * next2^;
  end else
  begin
   next := psingle(longint(pf) + 4);
   o2 := o1;
  end;
  if (cnt < 1) and (fpar05 > 0.5) then // interpolation?
   o1 := o1 * (1 - cnt) + cnt * next^; // get next sample

  cnt := cnt + (sr / samplerate) * 2 * fpar07; // increase samplecounter
  while (cnt >= 1) do
  begin
   inc(pf, ch);
   cnt := cnt - 1;
   inc(cnt2, ch);
  end;
  ls := round(fpar04 * (size - 1)); // calculate loop start
  le := round(fpar06 * (size - 1)); // calculate loop end
  if ls >= le then ls := 0;
  if (cnt2 >= le) then
  begin
   cnt2 := ls;
   cnt := 0;
   pf := psingle(longint(pbuf) + cnt2 * sizeof(single));
  end;
 end;
 if ch = 2 then // stereo output
 begin
  i1 := fpar01 * o1;
  i2 := fpar01 * o2;
 end else
 begin // mono output
  i1 := fpar01 * o1 * 2 * (1 - fpar02);
  i2 := fpar01 * o1 * 2 * fpar02;
 end;
end;

procedure APlugin.process(inputs, outputs: PPSingle; sampleframes: Longint);
var In1, In2, Out1, Out2: PSingle;
    i: Integer;
begin
 In1 := inputs^;
 In2 := PPSingle(Longint(inputs) + SizeOf(PSingle))^;
 Out1 := outputs^;
 Out2 := PPSingle(Longint(outputs) + SizeOf(PSingle))^;
 for i := 0 to sampleFrames - 1 do
 begin
  DoProcess(In1^, In2^);
  Out1^ := Out1^ + In1^;
  Out2^ := Out2^ + In2^;
  Inc(In1);
  Inc(In2);
  Inc(Out1);
  Inc(Out2);
 end;
end;

procedure APlugin.processReplacing(inputs, outputs: PPSingle; sampleframes: Longint);
var In1, In2, Out1, Out2: PSingle;
    i: Integer;
begin
 In1 := inputs^;
 In2 := PPSingle(Longint(inputs) + SizeOf(PSingle))^;
 Out1 := outputs^;
 Out2 := PPSingle(Longint(outputs) + SizeOf(PSingle))^;
 for i := 0 to sampleFrames - 1 do
 begin
  DoProcess(In1^, In2^);
  Out1^ := In1^;
  Out2^ := In2^;
  Inc(In1);
  Inc(In2);
  Inc(Out1);
  Inc(Out2);
 end;
end;

function APlugin.GetPluginEditor: APluginEditor;
begin
 Result := editor as APluginEditor;
end;

constructor APluginEditor.Create(effect: AudioEffect);
begin
 inherited Create(effect);
 useCount := 0;
end;

destructor APluginEditor.Destroy;
begin
 if assigned(editor) then
 begin
  Plugin.done := true;
  Editor.Free;
  Editor := nil;
  systemWindow := 0;
 end;
 inherited Destroy;
end;

function APluginEditor.getRect(var rect:PERect):Longint;
begin
 r.top := 0;
 r.left := 0;
 r.bottom := 447;
 r.right := 645;
 rect := @r;
 Result := 1;
end;

function APluginEditor.open(ptr: Pointer): Longint;
begin
 systemWindow := HWnd(ptr);
 Inc(useCount);
 if (useCount = 1) or (not assigned(editor)) then
 begin
  Editor := TPluginEditorWindow.CreateParented(systemWindow);
  Editor.SetBounds(0, 0, Editor.Width, Editor.Height);
  Editor.fpar01 := Plugin.fpar01;
  Editor.fpar02 := Plugin.fpar02;
  Editor.fpar03 := Plugin.fpar03;
  Editor.fpar04 := Plugin.fpar04;
  Editor.fpar05 := Plugin.fpar05;
  Editor.fpar06 := Plugin.fpar06;
  Editor.fpar07 := Plugin.fpar07;
  Editor.fpar08 := Plugin.fpar08;
  Editor.Effect := Self.Plugin;
 end;
 Plugin.done := false;
 Editor.Update;
 Editor.Show;
 Result := 1;
end;

procedure APluginEditor.close;
begin
 if assigned(editor) then editor.visible := false;
 Dec(useCount);
 if useCount < 0 then useCount := 0;
 if useCount = 0 then
 begin
  Plugin.done := true;
  Editor.Free;
  Editor := nil;
  systemWindow := 0;
 end;
end;

procedure APluginEditor.idle;
begin
 if (Plugin.done) or (not Assigned(Editor)) then exit;
 with editor do
 begin
  par01.position := round(fpar01 * 100);
  par02.position := round(fpar02 * 100);
  par04.position := round(fpar04 * 100);
  par06.position := round(fpar06 * 100);
  par07.position := round(fpar07 * 100);
  par05.checked := (fpar05 > 0.5);
 end;
end;

procedure APluginEditor.update;
begin
 if (Plugin.done) or (not Assigned(Editor)) then exit;
 Editor.fpar01 := Plugin.fpar01;
 Editor.fpar02 := Plugin.fpar02;
 Editor.fpar03 := Plugin.fpar03;
 Editor.fpar04 := Plugin.fpar04;
 Editor.fpar05 := Plugin.fpar05;
end;

function APluginEditor.GetPlugin: APlugin;
begin
 Result := (effect as APlugin);
end;

function APlugin.getEffectName(name: pchar): boolean;
begin
 StrCopy(name, kEffectName);
 Result := TRUE;
end;

function APlugin.getVendorString(text: pchar): boolean;
begin
 StrCopy(text, kVendor);
 Result := TRUE;
end;

function APlugin.getProductString(text: pchar):boolean;
begin
 StrCopy(text, kProduct);
 Result := TRUE;
end;

function APlugin.getVendorVersion: longint;
begin
 Result := 1; // return version number
end;

function APlugin.getOutputProperties(index: longint; properties: PVstPinProperties): boolean;
begin
 Result := false;
 if (index < kNumOutputs) then
 begin
  StrCopy(properties^.vLabel, pchar(Format(kChannelID+' %d', [index + 1])));
  properties^.flags := kVstPinIsActive;
  if (index < 2) then
   properties^.flags := properties^.flags or kVstPinIsStereo;
  Result := true;
 end;
end;

function APlugin.getProgramNameIndexed(category,index:longint;text:pchar):boolean;
begin
 Result := false;
 if (index < kNumPrograms) then
 begin
  StrCopy(text, programs[index].name);
  Result := true;
 end;
end;

procedure APlugin.getParameterName(index:longint;text:pchar);
begin
 StrCopy(text, 'undefined');
end;

procedure APlugin.getParameterDisplay(index:longint;text:pchar);
begin
 case index of
  kpar01: float2string(fpar01, text);
  kpar02: float2string(fpar02, text);
  kpar03: float2string(fpar03, text);
  kpar04: float2string(fpar04, text);
  kpar05: float2string(fpar05, text);
  kpar06: float2string(fpar06, text);
  kpar07: float2string(fpar07, text);
  kpar08: float2string(fpar08, text);
  else float2string(0, text);
 end;
end;

procedure APlugin.getParameterLabel(index:longint;text:pchar);
begin
 StrCopy(text, '%');
end;

function APlugin.processEvents(ev:PVstEvents):longint;
var k, i, note, status: longint;
    event: PVstMidiEvent;
    midiData: array[0..3] of byte;
begin
 for i := 0 to ev^.numEvents - 1 do
 if (ev.events[i].vtype = kVstMidiType) then
 begin
  event := PVstMidiEvent(ev^.events[i]);
  for k := 0 to 3 do midiData[k] := event.midiData[k];
  status := midiData[0] AND $f0; // channel information is removed
  if (status = $90) and (mididata[2] > 0) then // "note on" ?
  begin
   note := (midiData[1] and $7F); // midi note
   nvol := (midiData[2] and $7F) / 127; // velocity
  end
  else if ((status = $90) and (mididata[2] = 0)) or (status = $80) then // "note off" ?
  begin
   note := (midiData[1] and $7F); // midi note
  end
  else if (status = $B0) then // midi CC ?
  begin
   note := event^.mididata[1]; // midi CC#
   nvol := event^.mididata[2] / 127; // CC data
   if assigned(editor) then editor.update;
  end
  else if (status = $E0) then // pitchbend ?
  begin
   nvol := event^.mididata[1] / 127; // pitchbend data
   if assigned(editor) then editor.update;
  end;
 end;
 Result := 1;
end;

procedure APlugin.resume;
begin
 wantEvents(1); // important for all plugins that receive MIDI!
end;

end.

